from dataclasses import dataclass, field
from typing import Type

from data_to_paper.base_steps import DirectorProductGPT, DataStepRunner
from data_to_paper.latex.latex_doc import LatexDocument

from .app_startup import HypothesisTestingStartDialog
from .cast import ScientificAgent
from .coding.after_coding import RequestCodeExplanation, RequestCodeProducts
from .coding.analysis import DataAnalysisCodeProductsGPT
from .coding.exploration import DataExplorationCodeProductsGPT
from .coding.displayitems import CreateDisplayitemsCodeProductsGPT
from .coding.preprocessing import DataPreprocessingCodeProductsGPT
from .literature_search import WritingLiteratureSearchReviewGPT, GoalLiteratureSearchReviewGPT
from .produce_pdf_step import ProduceScientificPaperPDFWithAppendix
from .product_types import GoalAndHypothesisProduct
from .reviewing_steps import GoalReviewGPT, HypothesesTestingPlanReviewGPT, NoveltyAssessmentReview, ReGoalReviewGPT, \
    GetMostSimilarCitations
from .scientific_products import ScientificProducts
from .scientific_stage import ScientificStage
from .writing_steps import FirstTitleAbstractSectionWriterReviewGPT, SecondTitleAbstractSectionWriterReviewGPT, \
    MethodsSectionWriterReviewGPT, IntroductionSectionWriterReviewGPT, ResultsSectionWriterReviewGPT, \
    DiscussionSectionWriterReviewGPT

PAPER_SECTIONS_NAMES = ['title', 'abstract', 'introduction', 'results', 'discussion', 'methods']
SECTIONS_WITH_CITATIONS = ['introduction', 'discussion']

SKIP_STAGE_MESSAGE = 'This stage was skipped because the goal was provided by the user.'
SECTIONS_TO_WRITING_CLASS = [
    (('results',), ResultsSectionWriterReviewGPT),
    (('title', 'abstract'), SecondTitleAbstractSectionWriterReviewGPT),
    (('methods',), MethodsSectionWriterReviewGPT),
    (('introduction',), IntroductionSectionWriterReviewGPT),
    (('discussion',), DiscussionSectionWriterReviewGPT),
    # (('conclusion',), ConclusionSectionWriterReviewGPT),
]


@dataclass
class HypothesisTestingStepsRunner(DataStepRunner):
    PROJECT_PARAMETERS_FILENAME = 'data-to-paper-hypothesis-testing.json'
    DEFAULT_PROJECT_PARAMETERS = DataStepRunner.DEFAULT_PROJECT_PARAMETERS | dict(
        research_goal=None,
        should_do_data_exploration=True,
        should_do_data_preprocessing=False,
        should_prepare_hypothesis_testing_plan=True,
        should_do_literature_search=True,
        project_specific_goal_guidelines='',
        excluded_citation_titles=[],
        max_goal_refinement_iterations=3,
    )

    latex_document: LatexDocument = field(default_factory=LatexDocument)

    APP_STARTUP_CLS = HypothesisTestingStartDialog
    name = 'Hypothesis Testing Research'

    cast = ScientificAgent
    products: ScientificProducts = field(default_factory=ScientificProducts)
    stages: Type[ScientificStage] = ScientificStage

    goal_refinement_iteration: int = 0
    re_goal: bool = False

    def _pre_run_preparations(self):
        super()._pre_run_preparations()
        self.paper_producer = ProduceScientificPaperPDFWithAppendix.from_(
            self,
            output_filename='paper.pdf',
            paper_section_names=PAPER_SECTIONS_NAMES,
            figures_folder=self.output_directory,
        )

        self.stages_to_funcs = {
            ScientificStage.DATA: self._data_file_descriptions,
            ScientificStage.EXPLORATION: self._data_exploration,
            ScientificStage.GOAL: self._goal,
            ScientificStage.LITERATURE_REVIEW_GOAL: self._literature_search_goal,
            ScientificStage.ASSESS_NOVELTY: self._assess_novelty,
            ScientificStage.PLAN: self._hypotheses_testing_plan,
            ScientificStage.CODE: self._data_analysis,
            ScientificStage.DISPLAYITEMS: self._tables,
            ScientificStage.INTERPRETATION: self._interpretation,
            ScientificStage.LITERATURE_REVIEW_WRITING: self._literature_review_writing,
            ScientificStage.WRITING_RESULTS: self._writing_results,
            ScientificStage.WRITING_TITLE_AND_ABSTRACT: self._writing_title_and_abstract,
            ScientificStage.WRITING_METHODS: self._writing_methods,
            ScientificStage.WRITING_INTRODUCTION: self._writing_introduction,
            ScientificStage.WRITING_DISCUSSION: self._writing_discussion,
            ScientificStage.COMPILE: self._compile_paper,
        }

    """
    Stage functions
    """

    def _data_file_descriptions(self):
        self.director_converser = DirectorProductGPT.from_(
            self,
            assistant_agent=ScientificAgent.Director,
            user_agent=ScientificAgent.Performer,
            conversation_name='with_director',
        )
        self.products.data_file_descriptions = self.director_converser.get_product_or_no_product_from_director(
            product_name='Data description', returned_product=self.data_file_descriptions)
        self.send_product_to_client('data_file_descriptions')

    def _data_exploration(self):
        if not self.project_parameters['should_do_data_exploration']:
            return
        RequestCodeProducts.from_(
            self,
            code_step='data_exploration',
            code_writing_class=DataExplorationCodeProductsGPT,
            explain_code_class=RequestCodeExplanation,
            explain_created_files_class=None,
        ).get_code_and_output_and_descriptions()
        self.send_product_to_client('codes_and_outputs_with_explanations:data_exploration')

    def _literature_search_goal(self):
        if not self.project_parameters['should_do_literature_search']:
            return
        GoalLiteratureSearchReviewGPT.from_(
            self, excluded_citation_titles=self.project_parameters['excluded_citation_titles'],
            literature_search=self.products.literature_search['goal']
        ).get_literature_search()
        self.send_product_to_client('literature_search:goal')

    def _assess_novelty(self):
        self.products.most_similar_papers = GetMostSimilarCitations.from_(self).run_and_get_valid_result()
        self.products.novelty_assessment = NoveltyAssessmentReview.from_(self).run_and_get_valid_result()
        self.send_product_to_client('novelty_assessment')
        self.goal_refinement_iteration += 1
        if (not self.products.novelty_assessment['choice'] == 'OK' and
                self.goal_refinement_iteration < self.project_parameters['max_goal_refinement_iterations']):
            self.re_goal = True
            return ScientificStage.GOAL

    def _goal(self):
        next_stage = None
        if not self.re_goal:
            research_goal = self.director_converser.get_product_or_no_product_from_director(
                product_name='Research Goal', returned_product=self.project_parameters['research_goal'],
                acknowledge_no_product_message="OK. no problem. I will devise the goal myself.")
            if research_goal is None:
                # Goal not provided by the user
                self.products.research_goal = GoalReviewGPT.from_(
                    self,
                    project_specific_goal_guidelines=self.project_parameters['project_specific_goal_guidelines']
                ).run_and_get_valid_result()
            else:
                self.products.research_goal = GoalAndHypothesisProduct(value=research_goal)
                self._app_send_product_of_stage(ScientificStage.LITERATURE_REVIEW_GOAL, SKIP_STAGE_MESSAGE)
                self._app_send_product_of_stage(ScientificStage.ASSESS_NOVELTY, SKIP_STAGE_MESSAGE)
                next_stage = ScientificStage.PLAN
        else:
            self.products.research_goal = ReGoalReviewGPT.from_(
                self,
                project_specific_goal_guidelines=self.project_parameters['project_specific_goal_guidelines']
            ).run_and_get_valid_result()
        self.send_product_to_client('research_goal', save_to_file=True)
        return next_stage

    def _hypotheses_testing_plan(self):
        if not self.project_parameters['should_prepare_hypothesis_testing_plan']:
            return
        self.products.hypothesis_testing_plan = HypothesesTestingPlanReviewGPT.from_(
            self).run_and_get_valid_result()
        self.send_product_to_client('hypothesis_testing_plan', save_to_file=True)

    def _data_preprocessing(self):
        if not self.project_parameters['should_do_data_preprocessing']:
            return
        RequestCodeProducts.from_(
            self,
            code_step='data_preprocessing',
            code_writing_class=DataPreprocessingCodeProductsGPT,
            explain_code_class=RequestCodeExplanation,
            explain_created_files_class=None,
        ).get_code_and_output_and_descriptions()
        self.send_product_to_client('codes_and_outputs_with_explanations:data_preprocessing')

    def _data_analysis(self):
        RequestCodeProducts.from_(
            self,
            code_step='data_analysis',
            code_writing_class=DataAnalysisCodeProductsGPT,
            explain_code_class=RequestCodeExplanation,
            explain_created_files_class=None,
        ).get_code_and_output_and_descriptions()
        self.send_product_to_client('codes_and_outputs_with_explanations:data_analysis')

    def _tables(self):
        RequestCodeProducts.from_(
            self,
            code_step='data_to_latex',
            latex_document=self.latex_document,
            code_writing_class=CreateDisplayitemsCodeProductsGPT,
            explain_code_class=None,
            explain_created_files_class=None,
        ).get_code_and_output_and_descriptions()
        self.send_product_to_client('codes_and_outputs_with_explanations:data_to_latex')

    def _interpretation(self):
        self.products.paper_sections_and_optional_citations['title'], \
            self.products.paper_sections_and_optional_citations['abstract'] = \
            FirstTitleAbstractSectionWriterReviewGPT.from_(
                self, section_names=['title', 'abstract']).write_sections_with_citations()
        self.send_product_to_client('title_and_abstract_first')

    def _literature_review_writing(self):
        WritingLiteratureSearchReviewGPT.from_(
            self,
            literature_search=self.products.literature_search['writing'],
            excluded_citation_titles=self.project_parameters['excluded_citation_titles']
        ).get_literature_search()
        self.send_product_to_client('literature_search:writing')

    def _writing_paper_section(self, section_names, writing_class):
        sections_with_citations = \
            writing_class.from_(self, section_names=section_names).write_sections_with_citations()
        for section_name, section_and_citations in zip(section_names, sections_with_citations):
            self.products.paper_sections_and_optional_citations[section_name] = section_and_citations
        if len(section_names) == 2:
            self.send_product_to_client('title_and_abstract')
        else:
            self.send_product_to_client(f'paper_sections:{section_names[0]}')

    def _writing_results(self):
        self._writing_paper_section(('results',), ResultsSectionWriterReviewGPT)

    def _writing_title_and_abstract(self):
        self._writing_paper_section(('title', 'abstract'), SecondTitleAbstractSectionWriterReviewGPT)

    def _writing_methods(self):
        self._writing_paper_section(('methods',), MethodsSectionWriterReviewGPT)

    def _writing_introduction(self):
        self._writing_paper_section(('introduction',), IntroductionSectionWriterReviewGPT)

    def _writing_discussion(self):
        self._writing_paper_section(('discussion',), DiscussionSectionWriterReviewGPT)

    def _compile_paper(self):
        self.paper_producer.assemble_compile_paper()
        self._app_clear_panels()
        self._app_send_product_of_stage(
            ScientificStage.COMPILE,
            f'<a href="file://{self.output_directory}/paper.pdf">Download the manuscript</a>')



from __future__ import annotations
from dataclasses import dataclass, fields
from typing import List, Tuple, Dict, TYPE_CHECKING

from data_to_paper.utils.types import IndexOrderedEnum

if TYPE_CHECKING:
    from data_to_paper.servers.types import APIKey


class ModelEngine(IndexOrderedEnum):
    """
    Enum for the different model engines available.
    Support comparison operators, according to the order of the enum.
    """

    # ignore:
    _ignore_ = ['DEFAULT']

    DEFAULT = None
    GPT35_TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4"
    GPT4_TURBO = "gpt-4-turbo"
    GPT4o_MINI = "gpt-4o-mini"
    GPT4o = "gpt-4o"
    LLAMA_2_7b = "Llama-2-7b-chat-hf"
    LLAMA_2_70b = "Llama-2-70b-chat-hf"
    CODELLAMA = "CodeLlama-34b-Instruct-hf"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.value)

    def get_model_with_more_strength(self):
        model = MODELS_TO_MORE_STRENGTH.get(self)
        if model is None:
            raise ValueError(f"Model {self} has no stronger model")
        return model

    def get_model_with_more_context(self):
        model = MODELS_TO_MORE_CONTEXT.get(self)
        if model is None:
            raise ValueError(f"Model {self} has no model with more context")
        return model

    @property
    def max_tokens(self):
        return MODEL_ENGINE_TO_MAX_TOKENS_AND_IN_OUT_DOLLAR[self][0]

    @property
    def pricing(self) -> Tuple[float, float]:
        """
        Return the pricing for the model engine.
        (dollar_per_in_token, dollar_per_out_token)
        """
        return MODEL_ENGINE_TO_MAX_TOKENS_AND_IN_OUT_DOLLAR[self][1:]

    @property
    def allows_json_mode(self):
        return self in MODEL_ENGINES_ALLOWING_JSON_MODE

    @property
    def api_key(self) -> APIKey:
        return _get_api_key_and_server_name_and_base_url(self)[0]

    @property
    def base_url(self):
        return _get_api_key_and_server_name_and_base_url(self)[1]

    @property
    def server_name(self):
        return _get_api_key_and_server_name_and_base_url(self)[2]


ModelEngine.DEFAULT = ModelEngine.GPT4o_MINI

MODELS_TO_MORE_CONTEXT: Dict[ModelEngine, ModelEngine] = {
    ModelEngine.GPT35_TURBO: ModelEngine.GPT4_TURBO,
    ModelEngine.GPT4: ModelEngine.GPT4_TURBO,
    ModelEngine.GPT4_TURBO: ModelEngine.GPT4o,
    ModelEngine.GPT4o_MINI: ModelEngine.GPT4o,  # same as GPT4o
    ModelEngine.GPT4o: ModelEngine.GPT4o,  # same as GPT4o
}

MODELS_TO_MORE_STRENGTH: Dict[ModelEngine, ModelEngine] = {
    ModelEngine.GPT35_TURBO: ModelEngine.GPT4o,
    ModelEngine.GPT4: ModelEngine.GPT4o,
    ModelEngine.GPT4_TURBO: ModelEngine.GPT4o,
    ModelEngine.GPT4o_MINI: ModelEngine.GPT4o,
}

MODEL_ENGINE_TO_MAX_TOKENS_AND_IN_OUT_DOLLAR: Dict[ModelEngine, Tuple[int, float, float]] = {
    ModelEngine.GPT35_TURBO: (16384, 0.000001, 0.000002),
    ModelEngine.GPT4: (8192, 0.00003, 0.00006),
    ModelEngine.GPT4_TURBO: (128000, 0.00001, 0.00003),
    ModelEngine.GPT4o_MINI: (128000, 1.5e-7, 6e-7),
    ModelEngine.GPT4o: (128000, 0.000005, 0.000015),
    ModelEngine.LLAMA_2_7b: (4096, 0.0002, 0.0002),
    ModelEngine.LLAMA_2_70b: (4096, 0.0007, 0.001),
    ModelEngine.CODELLAMA: (4096, 0.0006, 0.0006),
}

MODEL_ENGINES_ALLOWING_JSON_MODE = {ModelEngine.GPT4o_MINI, ModelEngine.GPT4o}

OPENAI_API_BASE = "https://api.openai.com/v1"
DEEPINFRA_API_BASE = "https://api.deepinfra.com/v1/openai"


def _get_api_key_and_server_name_and_base_url(model_engine: ModelEngine) -> Tuple[APIKey, str, str]:
    from data_to_paper.env import OPENAI_API_KEY, DEEPINFRA_API_KEY
    open_ai_key_and_base_url = (OPENAI_API_KEY, OPENAI_API_BASE, "OpenAI")
    deep_infra_key_and_base_url = (DEEPINFRA_API_KEY, DEEPINFRA_API_BASE, "DeepInfra")

    llm_models_to_api_keys_and_base_url: Dict[ModelEngine, Tuple[APIKey, str, str]] = {
        ModelEngine.GPT35_TURBO: open_ai_key_and_base_url,
        ModelEngine.GPT4: open_ai_key_and_base_url,
        ModelEngine.GPT4_TURBO: open_ai_key_and_base_url,
        ModelEngine.GPT4o_MINI: open_ai_key_and_base_url,
        ModelEngine.GPT4o: open_ai_key_and_base_url,
        ModelEngine.LLAMA_2_7b: deep_infra_key_and_base_url,
        ModelEngine.LLAMA_2_70b: deep_infra_key_and_base_url,
        ModelEngine.CODELLAMA: deep_infra_key_and_base_url,
    }
    return llm_models_to_api_keys_and_base_url[model_engine]


@dataclass
class OpenaiCallParameters:
    """
    Parameters for calling OpenAI API.
    """
    model_engine: ModelEngine = None
    temperature: float = None
    max_tokens: int = None
    top_p: float = None
    frequency_penalty: float = None
    presence_penalty: float = None
    response_format: dict = None
    stop: List[str] = None

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def __str__(self):
        return str(self.to_dict())

    def is_all_none(self):
        return all(v is None for v in self.to_dict().values())


OPENAI_CALL_PARAMETERS_NAMES = [field.name for field in fields(OpenaiCallParameters)]



from typing import Iterable, Union, Optional, Any, List

import numpy as np
import pandas as pd

from data_to_paper.run_gpt_code.overrides.pvalue import is_p_value


def is_non_integer_numeric(value) -> bool:
    """
    Check if the value is a non-integer numeric.
    """
    if not isinstance(value, float):
        return False
    if is_p_value(value):
        return False
    if value.is_integer():
        return False
    # check if the value is nan or inf
    if np.isinf(value) or np.isnan(value):
        return False
    return True


def _find_longest_str_in_list(lst: Iterable[Union[str, Any]]) -> Optional[str]:
    """
    Find the longest string in a list of strings.
    Only iterate over the str labels.
    """
    try:
        longest_str = None
        length = 0
        for label in lst:
            if isinstance(label, str) and len(label) > length:
                longest_str = label
                length = len(label)
        return longest_str
    except ValueError:
        return None


def _find_longest_labels_in_index(index: [pd.Index, pd.MultiIndex]) -> List[str]:
    """
    Find the longest label in the index.
    For multi-index, return the longest label in each level.
    Should only iterate over the str labels.
    """
    if isinstance(index, pd.MultiIndex):
        return [_find_longest_str_in_list(index.get_level_values(level).unique()) for level in range(index.nlevels)]
    else:
        return [_find_longest_str_in_list(index)]


def _find_longest_labels_in_columns_relative_to_content(df, tolerance: int = 2) -> List[str]:
    """
    For each column we need to check to what extent its label extends its content width (the width it would have if
    there was no label)
    """
    if isinstance(df.columns, pd.MultiIndex):
        # TODO: Implement for multi-index
        return []
    column_label_widths = np.array([len(label) for label in df.columns])
    column_content_widths = np.zeros_like(column_label_widths)
    for i, column in enumerate(df.columns):
        column_content_widths[i] = max(len(str(value)) for value in df[column])
    return df.columns[column_label_widths - column_content_widths > tolerance]



import importlib

KNOWN_MIS_IMPORTS = {
    'Mediation': 'statsmodels.stats.mediation',
}


# assert KNOWN_MIS_IMPORTS:
def _assert_known_mis_imports():
    for name, module in KNOWN_MIS_IMPORTS.items():
        try:
            importlib.import_module(module, name)
        except ImportError:
            raise ImportError(f"Wrong imports in KNOWN_MIS_IMPORTS.\nFailed importing {name} from {module}")


_assert_known_mis_imports()



import importlib
import os
import re
import shutil
import subprocess
import fitz  # PyMuPDF
import numpy as np

from typing import Optional, Collection, Tuple, Dict

from pathlib import Path

from data_to_paper.utils.subprocess_call import get_subprocess_kwargs
from data_to_paper.terminate.exceptions import MissingInstallationError
from data_to_paper.servers.custom_types import Citation
from data_to_paper.utils.file_utils import run_in_temp_directory
from data_to_paper.code_and_output_files.ref_numeric_values import replace_hyperlinks_with_values
from data_to_paper.text.text_extractors import extract_all_external_brackets

from .exceptions import LatexCompilationError, LatexNumCommandFormulaEvalError, \
    LatexNestedNumCommandError, LatexNumCommandNoExplanation, PlainNumberLatexNumCommandError

BIB_FILENAME: str = 'citations.bib'

PDFLATEX_INSTALLATION_INSTRUCTIONS = r"""
Installations instructions for pdflatex:

* **On Ubuntu**:
```
sudo apt-get update && \
sudo apt-get install -y --no-install-recommends \
texlive-latex-base \
texlive-latex-extra \
texlive-fonts-recommended
```

* **On MacOS**:
- Ensure you have Homebrew installed (see https://brew.sh).
- Install MacTeX with the following command:
```
brew install --cask mactex-no-gui
```
- After installation, you may need to add the TeX binaries to your PATH!

* **On Windows**:
- Download and install MiKTeX from https://miktex.org/download.
- During installation, select 'Yes' when asked to install missing packages on-the-fly.
- After installation, you may need to add the TeX binaries to your PATH!
 """


def is_pdflatex_installed() -> Optional[bool]:
    """
    Check that pdflatex is installed.
    True if installed, False if not.
    None if check failed.
    """
    try:
        subprocess.run(['pdflatex', '--version'], **get_subprocess_kwargs(capture=False))
    except FileNotFoundError:
        return False
    except subprocess.CalledProcessError:
        return None
    return True


def is_pdflatex_package_installed(package: str) -> Optional[bool]:
    """
    Check that the packages used in the latex document are installed.
    True if installed, False if not.
    None if check failed.
    """
    try:
        subprocess.run(['kpsewhich', package + '.sty'], **get_subprocess_kwargs(capture=False))
    except FileNotFoundError:
        return None
    except subprocess.CalledProcessError:
        return False
    return True


def is_string_plain_number(string: str) -> bool:
    try:
        float(string)
    except ValueError:
        return False
    return True


def evaluate_latex_num_command(latex_str, ref_prefix='', enforce_explanation: bool = True,
                               just_strip_explanation: bool = False,
                               ) -> Tuple[str, Dict[str, str]]:
    r"""
    Evaluates all expressions of the form \num{formula} or \num{formula, "explanation"} in the given latex string
    and replaces them with the result of the formula.
    if ref_prefix, then add \hyperlink{ref_prefix?}{result}, where ? is the index of the expression.
    Return the new latex string and a mapping from the index to "expression = result".
    """
    command = r'\num{'
    matches = extract_all_external_brackets(latex_str, '{', '}', open_phrase=command)
    labels_to_notes = {}
    for index, full_match in enumerate(matches):
        match = full_match[len(command):-1]  # remove the command and the closing bracket
        match = match.strip()
        if command in match:
            raise LatexNestedNumCommandError(expression=full_match)
        # separate <formula>, "explanation" into formula and explanation:
        if match.endswith('"'):
            open_quote_index = match[:-1].rfind('"')
            comma_index = match[:open_quote_index].rfind(',')
            explanation = match[open_quote_index + 1:-1]
            formula = match[:comma_index]
        else:
            explanation = None
            formula = match

        formula_without_hyperlinks = replace_hyperlinks_with_values(formula)
        if is_string_plain_number(formula_without_hyperlinks):
            raise PlainNumberLatexNumCommandError(expression=full_match)
        if enforce_explanation and explanation is None:
            raise LatexNumCommandNoExplanation(expression=full_match)
        if just_strip_explanation:
            replace_with = command + formula_without_hyperlinks + '}'
            latex_str = latex_str.replace(command + match + '}', replace_with)
            continue
        try:
            result = eval(formula_without_hyperlinks,
                          {'exp': np.exp, 'log': np.log, 'sin': np.sin, 'cos': np.cos, 'tan': np.tan, 'pi': np.pi,
                           'e': np.e, 'sqrt': np.sqrt, 'log2': np.log2, 'log10': np.log10})
        except Exception as e:
            raise LatexNumCommandFormulaEvalError(expression=match, exception=e)
        if isinstance(result, float):
            result = '{:.4g}'.format(result)
        else:
            result = str(result)
        label = f'{ref_prefix}{index}'
        if ref_prefix:
            replace_with = f'\\hyperlink{{{label}}}{{{result}}}'
        else:
            replace_with = result
        latex_str = latex_str.replace(command + match + '}', replace_with)
        note = f'{formula} = {result}'
        if explanation:
            note += f'\n\n{explanation}'
            # note += f'     {explanation}'
        labels_to_notes[label] = note
    return latex_str, labels_to_notes


def add_watermark_to_pdf(pdf_path: str, watermark_path: str, output_path: str = None):
    """
    Add watermark to each page of a PDF while preserving hyperlinks using PyMuPDF (fitz).
    :param pdf_path: Path to the PDF file to be watermarked.
    :param watermark_path: Path to the watermark PDF file. The first page of this file will be used as the watermark.
    :param output_path: Path for the output watermarked PDF file. If None, the original PDF will be overwritten.
    """
    if output_path is None:
        output_path = pdf_path

    with fitz.open(pdf_path) as pdf_doc, fitz.open(watermark_path) as watermark_doc:
        for page in pdf_doc:
            # Overlay the watermark onto the page by adding the XObject
            page.show_pdf_page(page.rect, watermark_doc, 0)

        # Save the watermarked PDF
        pdf_doc.save(output_path, incremental=True, encryption=0)


def _get_over_width_pts(pdflatex_output: str) -> Optional[float]:
    match = re.search(pattern=r'Overfull \\hbox \((.*?)pt too wide\)', string=pdflatex_output)
    if match:
        return float(match.group(1))
    return None


def save_latex_and_compile_to_pdf(latex_content: str, file_stem: str, output_directory: Optional[str] = None,
                                  references: Collection[Citation] = None, format_cite: bool = True,
                                  figures_folder: Optional[Path] = None,
                                  ) -> Tuple[str, Optional[float]]:
    references = references or set()
    should_compile_with_bib = len(references) > 0
    latex_file_name = file_stem + '.tex'
    pdflatex_params = ['pdflatex', '--shell-escape', '-interaction=nonstopmode', latex_file_name]
    with run_in_temp_directory():
        # Copy the figures from the output directory to the temp directory:
        if figures_folder is not None:
            png_files_in_running_directory = [f for f in figures_folder.glob('*.png') if f.is_file()]
            for png_file in png_files_in_running_directory:
                shutil.copy(png_file, '.')

        # Create the bib file:
        if should_compile_with_bib:
            references_bibtex = [reference.bibtex for reference in references]
            with open(BIB_FILENAME, 'w', encoding='utf-8') as f:
                f.write('\n\n'.join(references_bibtex))

        with open(latex_file_name, 'w', encoding='utf-8') as f:
            f.write(latex_content)
        try:
            pdflatex_output = subprocess.run(pdflatex_params, **get_subprocess_kwargs())
        except FileNotFoundError:
            raise MissingInstallationError(package_name="pdflatex", instructions=PDFLATEX_INSTALLATION_INSTRUCTIONS)
        except subprocess.CalledProcessError as e:
            _move_latex_and_pdf_to_output_directory(file_stem, output_directory, latex_file_name)
            raise LatexCompilationError(latex_content=latex_content,
                                        pdflatex_output=e.stdout.decode('utf-8', errors='replace'))

        pdflatex_output = pdflatex_output.stdout.decode('utf-8', errors='replace')

        if format_cite:
            try:
                if should_compile_with_bib:
                    try:
                        subprocess.run(['bibtex', file_stem], **get_subprocess_kwargs(capture=False))
                    except FileNotFoundError:
                        raise MissingInstallationError(package_name="bibtex",
                                                       instructions=PDFLATEX_INSTALLATION_INSTRUCTIONS)
                subprocess.run(pdflatex_params, **get_subprocess_kwargs(capture=False))
                subprocess.run(pdflatex_params, **get_subprocess_kwargs(capture=False))
            except subprocess.CalledProcessError:
                _move_latex_and_pdf_to_output_directory(file_stem, output_directory, latex_file_name)
                raise

        with importlib.resources.path('data_to_paper.latex.resources', 'watermark.pdf') as watermark_path:
            add_watermark_to_pdf(file_stem + '.pdf', str(watermark_path))

        _move_latex_and_pdf_to_output_directory(file_stem, output_directory, latex_file_name)
        over_width_pts = _get_over_width_pts(pdflatex_output)
        return pdflatex_output, over_width_pts


def _move_latex_and_pdf_to_output_directory(file_stem: str, output_directory: str = None, latex_file_name: str = None):
    # Move the pdf and the latex and the citation file to the original directory:

    def move_if_exists(file_name):
        if os.path.exists(file_name):
            # delete older file if exists (this happen when we reset from the compilation step to earlier step)
            file_path = os.path.join(output_directory, file_name)
            if os.path.exists(file_path):
                os.remove(file_path)
            shutil.move(file_name, output_directory)

    if output_directory is not None:
        move_if_exists(file_stem + '.pdf')
        move_if_exists(latex_file_name)
        move_if_exists('citations.bib')



from __future__ import annotations
import os
import pickle
from dataclasses import dataclass, field
from fnmatch import fnmatch
from pathlib import Path
from typing import Optional, Any, List, Tuple, Iterable, Dict, Type

from data_to_paper.env import MAX_SENSIBLE_OUTPUT_SIZE_TOKENS, NUM_DIGITS_FOR_FLOATS
from data_to_paper.servers.llm_call import count_number_of_tokens_in_message
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.text import dedent_triple_quote_str, wrap_as_block
from data_to_paper.text.highlighted_text import format_text_with_code_blocks
from data_to_paper.text.text_extractors import extract_to_nearest_newline
from data_to_paper.text.text_numeric_formatting import round_floats
from data_to_paper.code_and_output_files.referencable_text import NumericReferenceableText, \
    BaseReferenceableText, convert_str_to_latex_label

from data_to_paper.run_gpt_code.overrides.pvalue import OnStrPValue, OnStr
from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue

from .file_view_params import ContentViewPurposeConverter, ViewPurpose
from .ref_numeric_values import ReferencedValue, HypertargetFormat

"""
OUTPUT FILEE REQUIREMENTS
-------------------------
There are two main types of requirements:

- Data: the file contains data that is not presentable to the LLM. Like: large data files, images, etc.
In this case we keep the file.

- Content: the file contains content that is presentable to the LLM. Like: text, tables, etc.
In this case we load and keep the content and typically delete the file.
"""

EXTS_TO_LABELS = {
    '.tex': 'latex',
    '.txt': 'output',
    '.csv': 'csv',
}


@dataclass(frozen=True)
class OutputFileRequirement:
    generic_filename: str
    minimal_count: int
    should_keep_file: bool = NotImplemented
    should_make_available_for_next_steps: bool = True

    def is_wildcard(self):
        return '*' in self.generic_filename or '?' in self.generic_filename

    def matches(self, filename: str):
        return fnmatch(filename, self.generic_filename)

    def delete_if_needed(self, file_path: str):
        """
        Delete the file if needed.
        """
        if not self.should_keep_file:
            os.remove(file_path)

    def read_content(self, file_path: str) -> Optional[str]:
        """
        Return the content of the file.
        If data file, return None.
        """
        return None

    def get_content_and_delete_if_needed(self, file_path: str) -> str:
        """
        Return the content of the file, and delete it if needed.
        """
        content = self.read_content(file_path)
        self.delete_if_needed(file_path)
        return content

    """Linking"""

    def get_code_line_str_for_file(self, filename: str, content: Any) -> Optional[str]:
        """
        Return a string which can be found in the line where we should go to when we want to see the code
        that created the file.
        """
        return filename

    def get_code_line_label_for_file(self, filename: str, content: Optional[Any] = None) -> Optional[str]:
        """
        Return a label to add to the code at the line where we should go to when we want to see the code
        """
        return convert_str_to_latex_label(filename, prefix='code')

    def get_hyperlink_label_for_file_header(self, filename: str, content: Any) -> Optional[str]:
        """
        Return a hypertarget label to go to when we click the file header.
        None to not create a hyperlink.
        """
        return self.get_code_line_label_for_file(filename, content)


@dataclass(frozen=True)
class DataOutputFileRequirement(OutputFileRequirement):
    minimal_count: int = 0
    should_keep_file: bool = True


@dataclass(frozen=True)
class BaseContentOutputFileRequirement(OutputFileRequirement):
    should_keep_file: bool = NotImplemented
    minimal_count: int = 1
    content_view_purpose_converter: ContentViewPurposeConverter = field(default_factory=ContentViewPurposeConverter)
    VIEW_PURPOSE_TO_PVALUE_ON_STR = {
        ViewPurpose.PRODUCT: OnStr.SMALLER_THAN,
        ViewPurpose.HYPERTARGET_PRODUCT: OnStr.SMALLER_THAN,
        ViewPurpose.APP_HTML: OnStr.WITH_ZERO,
        ViewPurpose.CODE_REVIEW: OnStr.WITH_EPSILON,
        ViewPurpose.FINAL_APPENDIX: OnStr.WITH_ZERO,
        ViewPurpose.FINAL_INLINE: OnStr.LATEX_SMALLER_THAN,
    }

    """READ AND CHECK FILE"""

    def read_content(self, file_path: str) -> Any:
        """
        Return the content of the file.
        """
        with open(file_path, 'r') as file:
            return file.read()

    def get_issues_for_output_file_content(self, filename: str, content: Any) -> List[RunIssue]:
        """
        Check the output and return a list of issues.
        """
        return []

    """VIEW FILE"""

    def _convert_view_purpose_to_pvalue_on_str(self, view_purpose: ViewPurpose) -> OnStr:
        return self.VIEW_PURPOSE_TO_PVALUE_ON_STR[view_purpose]

    def _to_str(self, content: Any, filename: str = None, num_file: int = 0, view_purpose: ViewPurpose = None) -> str:
        pvalue_on_str = self._convert_view_purpose_to_pvalue_on_str(view_purpose)
        with OnStrPValue(pvalue_on_str):
            return str(content).strip()

    def _get_block_label(self, filename: str, num_file: int, view_purpose: ViewPurpose) -> str:
        """
        Return the block label for the filename.
        """
        ext = os.path.splitext(filename)[1]
        return EXTS_TO_LABELS.get(ext, 'output')

    def get_header(self, filename: str = None, num_file: int = 0,
                   view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
                   header_level: int = 3) -> str:
        return self._get_pretty_content_and_header(None, filename, num_file, view_purpose, header_level)[1]

    def get_pretty_content(self, content: Any, filename: str = None, num_file: int = 0,
                           view_purpose: ViewPurpose = ViewPurpose.PRODUCT) -> str:
        return self._get_pretty_content_and_header(content, filename, num_file, view_purpose)[0]

    def get_pretty_content_with_header(self, content: Any, filename: str = None, num_file: int = 0,
                                       view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
                                       header_level: Optional[int] = 3) -> str:
        content, header = self._get_pretty_content_and_header(content, filename, num_file, view_purpose, header_level)
        if header_level:
            newline = '<br>\n' if view_purpose.is_for_html() else '\n'
            if header:
                return header + newline + content
        return content

    def _get_pretty_content_and_header(self, content: Any, filename: str = None, num_file: int = 0,
                                       view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
                                       header_level: Optional[int] = 3) -> Tuple[str, str]:
        purpose_to_func = {
            ViewPurpose.APP_HTML: self._get_content_and_header_for_app_html,
            ViewPurpose.PRODUCT: self._get_content_and_header_for_product,
            ViewPurpose.HYPERTARGET_PRODUCT: self._get_content_and_header_for_product,
            ViewPurpose.CODE_REVIEW: self._get_content_and_header_for_code_review,
            ViewPurpose.FINAL_APPENDIX: self._get_content_and_header_for_final_appendix,
            ViewPurpose.FINAL_INLINE: self._get_content_and_header_for_final_inline,
        }
        func = purpose_to_func[view_purpose]
        return func(content, filename, num_file, header_level, view_purpose)

    """Override these methods in subclasses to customize the output file content."""

    def _get_content_and_header_for_app_html(
            self, content: Any, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.APP_HTML):
        content = self._to_str(content, filename, num_file, view_purpose)
        content = format_text_with_code_blocks(content, from_md=True, is_html=True, width=None)
        return content, f'<h{level}>{filename}</h{level}>'

    def _get_content_and_header_for_product(
            self, content: Any, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.PRODUCT):
        content = self._to_str(content, filename, num_file, view_purpose)
        content = wrap_as_block(content, self._get_block_label(filename, num_file, view_purpose))
        return content, '#' * level + f' {filename}'

    def _get_content_and_header_for_code_review(
            self, content: Any, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.CODE_REVIEW):
        return self._get_content_and_header_for_product(content, filename, num_file, level, view_purpose)

    def _get_content_and_header_for_final_appendix(
            self, content: Any, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.FINAL_APPENDIX):
        return self._to_str(content, filename, num_file, view_purpose), f'% {filename}'

    def _get_content_and_header_for_final_inline(
            self, content: Any, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.FINAL_INLINE):
        return self._to_str(content, filename, num_file, view_purpose), f'% {filename}'


@dataclass(frozen=True)
class ReferencableContentOutputFileRequirement(BaseContentOutputFileRequirement):
    hypertarget_prefixes: Optional[Tuple[str]] = None  # List of hypertarget prefixes to assign for each file
    referenceable_text_cls: Type[BaseReferenceableText] = NumericReferenceableText

    def _to_str(self, content: Any, filename: str = None, num_file: int = 0, view_purpose: ViewPurpose = None) -> str:
        return self.get_formatted_text_and_header_references(content, filename, num_file, view_purpose)[0]

    def _get_prefix(self, num_file: int) -> str:
        return self.hypertarget_prefixes[num_file] if self.hypertarget_prefixes else None

    def get_formatted_text_and_header_references(self, content: Any, filename: str = None, num_file: int = 0,
                                                 view_purpose: ViewPurpose = None) -> Tuple[str, List[ReferencedValue]]:
        referencable_text = self._get_referencable_text(content, filename, num_file, view_purpose)
        return referencable_text.get_formatted_text_and_header_references(
            hypertarget_format=self._get_hyper_target_format(content, filename, num_file, view_purpose))

    def _convert_content_to_labeled_text(self, content: Any, filename: str = None, num_file: int = 0,
                                         view_purpose: ViewPurpose = None) -> str:
        return super()._to_str(content, filename, num_file, view_purpose)

    def _get_referencable_text(self, content: Any, filename: str = None, num_file: int = 0,
                               view_purpose: ViewPurpose = None) -> BaseReferenceableText:
        return self.referenceable_text_cls(
            text=self._convert_content_to_labeled_text(content, filename, num_file, view_purpose),
            hypertarget_prefix=self._get_prefix(num_file),
        )

    def _get_hyper_target_format(self, content: Any, filename: str = None, num_file: int = 0,
                                 view_purpose: ViewPurpose = None) -> HypertargetFormat:
        # prefix = self._get_prefix(num_file)
        # if prefix is None:
        #     return HypertargetFormat()  # no hypertargets
        return self.content_view_purpose_converter.convert_view_purpose_to_hypertarget_format(view_purpose)


@dataclass(frozen=True)
class PickleContentOutputFileRequirement(ReferencableContentOutputFileRequirement):
    should_keep_file: bool = True

    def read_content(self, file_path: str) -> Any:
        """
        Return the content of the file.
        """
        with open(file_path, 'rb') as file:
            return pickle.load(file)


@dataclass(frozen=True)
class TextContentOutputFileRequirement(ReferencableContentOutputFileRequirement):
    should_keep_file: bool = False
    max_tokens: Optional[int] = MAX_SENSIBLE_OUTPUT_SIZE_TOKENS.val

    def get_issues_for_output_file_content(self, filename: str, content: str) -> List[RunIssue]:
        issues = super().get_issues_for_output_file_content(filename, content)

        if len(content.strip()) == 0:
            # The output file is empty.
            issues.append(RunIssue(
                category='Problem in output file(s)',
                item=filename,
                issue=f'The code created the output file "{filename}", but the file is just empty!',
                instructions="Please revise the code to make sure it correctly writes to the output file.",
                code_problem=CodeProblem.OutputFileContentA,
            ))

        if self.max_tokens is not None \
                and count_number_of_tokens_in_message(content, max(ModelEngine)) > self.max_tokens:
            # Created output file is too large.
            issues.append(RunIssue(
                category='Problem in output file(s)',
                item=filename,
                issue=dedent_triple_quote_str("""
                    The code created the output file "{}", but the file is too long!

                    Here, for context, is the beginning of the output:
                    ```output
                    {}
                    ```
                    """).format(filename, extract_to_nearest_newline(content, self.max_tokens)),
                instructions="Only sensible-length output should be written to the file.",
                code_problem=CodeProblem.OutputFileContentB,
            ))

        return issues


@dataclass(frozen=True)
class NumericTextContentOutputFileRequirement(ReferencableContentOutputFileRequirement):
    target_precision: int = NUM_DIGITS_FOR_FLOATS
    source_precision: int = 10

    def _to_str(self, content: Any, filename: str = None, num_file: int = 0, view_purpose: ViewPurpose = None) -> str:
        content = super()._to_str(content, filename, num_file, view_purpose)
        return round_floats(content, self.target_precision, self.source_precision)


class OutputFileRequirements(Tuple[OutputFileRequirement]):
    """
    Stores a list of requirements for the output files of an LLM code run.
    """

    def get_all_allowed_created_filenames(self) -> Tuple[str, ...]:
        return tuple(requirement.generic_filename for requirement in self)

    def _match_files_to_requirements(self, created_files: Iterable[str]) -> Dict[str, Optional[OutputFileRequirement]]:
        """
        Return a dictionary mapping each file to the requirement that it matches.
        If a file does not match any requirement, the value is None.
        """
        file_to_requirement = {}
        for created_file in created_files:
            for requirement in self:
                if requirement.matches(created_file):
                    file_to_requirement[created_file] = requirement
                    break
            else:
                file_to_requirement[created_file] = None
        return file_to_requirement

    def get_requirements_to_output_files(
            self, created_files: Iterable[str]) -> Dict[OutputFileRequirement, List[str]]:
        """
        Return a dictionary mapping each requirement to a list of output files that match it.
        """
        requirements_to_files = {requirement: [] for requirement in self}
        for created_file, requirement in self._match_files_to_requirements(created_files).items():
            if requirement is not None:
                requirements_to_files[requirement].append(created_file)
        return requirements_to_files

    def get_unmatched_files(self, created_files: Iterable[str]) -> List[str]:
        return [created_file for created_file, requirement in self._match_files_to_requirements(created_files).items()
                if requirement is None]

    def convert_to_output_file_requirements_with_content(self, created_files: Iterable[str],
                                                         run_folder) -> OutputFileRequirementsToFileToContent:
        """
        Returns an OutputFileRequirementsToFileToContent, which is a dictionary mapping each requirement to
        a dictionary mapping each output file to its content.
        """
        requirements_to_files = self.get_requirements_to_output_files(sorted(created_files))
        requirements_to_files_to_content = \
            {requirement: {
                output_file: requirement.get_content_and_delete_if_needed(
                    file_path=run_folder / output_file if run_folder else output_file)
                for output_file in files
            } for requirement, files in requirements_to_files.items()}
        return OutputFileRequirementsToFileToContent(requirements_to_files_to_content)


class OutputFileRequirementsToFileToContent(Dict[OutputFileRequirement, Dict[str, Any]]):
    """
    A dictionary mapping each requirement to a dictionary mapping each output file to its content.
    Allows to get textual content of the files.
    """

    def convert_to_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements(self.keys())

    def get_all_created_files(self) -> List[str]:
        """
        Return the names of all the files created by the run.
        """
        return [filename for filenames_to_contents in self.values() for filename in filenames_to_contents.keys()]

    def get_created_files_available_for_next_steps(self) -> List[str]:
        """
        Return the names of the files created by the run, and which should be made available for the next steps.
        """
        return [filename for requirement, files_to_contents in self.items()
                for filename in files_to_contents.keys() if
                requirement.should_keep_file and requirement.should_make_available_for_next_steps]

    def get_created_files_to_requirements_and_contents(self, match_filename: str = '*',
                                                       is_content: bool = None,
                                                       ) -> Dict[str, Tuple[OutputFileRequirement, Any]]:
        """
        Return the names of the files created by the run, and their requirements and content.
        Content is `None` for data files.
        is_content:
            True: return only content files
            False: return only data files
            None: return all files
        """
        return {filename: (requirement, content) for requirement, files_to_contents in self.items()
                for filename, content in files_to_contents.items() if fnmatch(filename, match_filename)
                and (is_content is None or is_content == (content is not None))}

    def get_created_content_files_to_contents(self, match_filename: str = '*') -> Dict[str, Any]:
        """
        Return the names of the files created by the run, and their content.
        """
        return {filename: content for filename, (requirement, content) in
                self.get_created_files_to_requirements_and_contents(match_filename, is_content=True).items()}

    def get_created_content_files(self, match_filename: str = '*') -> List[str]:
        """
        Return the names of the files created by the run, for which we collected the content.
        """
        return list(self.get_created_content_files_to_contents(match_filename).keys())

    def get_created_data_files(self, match_filename: str = '*') -> List[str]:
        """
        Return the names of the files created by the run, and which were kept, not deleted.
        """
        return list(self.get_created_files_to_requirements_and_contents(match_filename, is_content=False).keys())

    def delete_all_created_files(self, run_folder: Optional[Path] = None):
        """
        Delete all the files that were created by the run, and which were kept, not deleted.
        """
        for filename in self.get_all_created_files():
            filepath = run_folder / filename if run_folder else filename
            if os.path.exists(filepath):
                os.remove(filepath)

    def get_created_content_files_to_pretty_contents(self, view_purpose: ViewPurpose,
                                                     match_filename: str = '*',
                                                     header_level: Optional[int] = 3) -> Dict[str, str]:
        """
        Return the names of the files created by the run, and their content formatted for display.
        """
        files_to_pretty_contents = {}
        for requirement, files_to_contents in self.items():
            for num_file, (filename, content) in enumerate(files_to_contents.items()):
                if isinstance(requirement, BaseContentOutputFileRequirement) and fnmatch(filename, match_filename):
                    pretty_content = requirement.get_pretty_content_with_header(
                        content, filename, num_file, view_purpose, header_level=header_level)
                    files_to_pretty_contents[filename] = pretty_content

        return files_to_pretty_contents

    def get_created_content_files_and_contents_as_single_str(self, view_purpose: ViewPurpose,
                                                             match_filename: str = '*',
                                                             header_level: Optional[int] = 3) -> str:
        files_to_contents = self.get_created_content_files_to_pretty_contents(
            view_purpose=view_purpose, match_filename=match_filename, header_level=header_level)
        return '\n\n'.join(files_to_contents.values())



import functools
import inspect
from dataclasses import dataclass
from typing import Iterable, Optional

from data_to_paper.env import TRACK_P_VALUES
from data_to_paper.run_gpt_code.attr_replacers import SystematicFuncReplacerContext

from ..pvalue import convert_to_p_value, TrackPValueCreationFuncs
from ..types import is_namedtuple, NoIterTuple
from ..utils import get_func_call_str


@dataclass
class ScipyPValueOverride(SystematicFuncReplacerContext, TrackPValueCreationFuncs):
    prevent_unpacking: Optional[bool] = True  # False - do not prevent;  True - prevent;  None - create issues
    package_names: Iterable[str] = ('scipy', )
    obj_import_str: str = 'scipy'

    def _should_replace(self, module, func_name, func) -> bool:
        doc = inspect.getdoc(func)
        if doc and "p-value" in doc:
            return True
        return False

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(*args, **kwargs):
            result = original_func(*args, **kwargs)
            created_by = original_func.__name__

            if TRACK_P_VALUES:
                # Get function call string representation:
                # For each arg in args, get a short representation of it, like 'array(shape=(2, 3))':
                func_call_str = get_func_call_str(created_by, args, kwargs)
                # Replace the pvalues attribute if it exists
                try:
                    asdict = {k.strip('_'): v for k, v in result._asdict().items()}
                    if 'pvalue' in asdict:
                        asdict['pvalue'] = convert_to_p_value(asdict['pvalue'],
                                                              created_by=created_by,
                                                              func_call_str=func_call_str)
                        self._add_pvalue_creating_func(created_by)
                        result = type(result)(**asdict)
                except (AttributeError, TypeError, ValueError):
                    pass

            if self.prevent_unpacking is not False and is_namedtuple(result):
                result = NoIterTuple(result, created_by=created_by, context=self,
                                     should_raise=self.prevent_unpacking is True,
                                     should_record=self.prevent_unpacking is None)

            return result

        return wrapped



from dataclasses import dataclass
from typing import Tuple, Optional

from data_to_paper.text import dedent_triple_quote_str

from .base_products_conversers import BackgroundProductsConverser
from .result_converser import Rewind


@dataclass
class MultiChoiceBackgroundProductsConverser(BackgroundProductsConverser):
    """
    A base class for asking LLM to choose between multiple options.
    """

    LLM_PARAMETERS = {'temperature': 0.5, 'max_tokens': 30}

    mission_prompt: str = dedent_triple_quote_str("""
        Please choose one of the following options:
        1. Looks good. Choice 1.
        2. Something is wrong. Choice 2.

        Your response should be formatted as {your_response_should_be_formatted_as}
        """)

    your_response_should_be_formatted_as: str = \
        "a single character, designating the option you choose {possible_choices}."

    possible_choices: Tuple[str, ...] = ('1', '2')

    default_rewind_for_result_error: Rewind = Rewind.AS_FRESH
    rewind_after_getting_a_valid_response: Optional[Rewind] = Rewind.AS_FRESH

    def _get_chosen_choice_from_response(self, response: str) -> str:
        choices_in_response = [choice for choice in self.possible_choices if choice in response]
        if len(choices_in_response) == 1:
            return choices_in_response[0]
        self._raise_self_response_error(
            title='# Could not extract the chosen choice',
            error_message=f'Could not extract a single choice from the response.',
        )

    def _check_extracted_text_and_update_valid_result(self, extracted_text: str):
        chosen_choice = self._get_chosen_choice_from_response(extracted_text)
        self._update_valid_result(chosen_choice)



from .conversation import Conversation
from .conversation_manager import ConversationManager
from .message import Role, Message
from .message_designation import GeneralMessageDesignation
from .conversation_actions import Action



# pragma: no cover
"""
Obsolete. We are now enforcing removal of min/max in rule-based review.
"""

from typing import Optional
from pandas.core.frame import DataFrame


class ModifiedDescribeDF(DataFrame):
    def _drop_rows(self, drop_count: Optional[bool] = False):
        to_drop = ['min', '25%', '50%', '75%', 'max']
        if drop_count or drop_count is None and all(self.loc['count'] == self.loc['count'][0]):
            # if all counts are the same, we drop the count row
            to_drop.append('count')
        return self.drop(to_drop)

    def __str__(self):
        return DataFrame.__str__(self._drop_rows())

    def __repr__(self):
        return DataFrame.__repr__(self._drop_rows())

    def to_string(self, *args, **kwargs):
        return DataFrame.to_string(self._drop_rows(), *args, **kwargs)

    # def to_latex(self, *args, **kwargs):
    #     return DataFrame.to_latex(self._drop_rows(drop_count=None), *args, **kwargs)


def describe(self, *args, original_method=None, on_change=None, **kwargs):
    """
    Removes the min, 25%, 50%, 75%, max rows from the result of the original describe function.
    """
    result = original_method(self, *args, **kwargs)
    return ModifiedDescribeDF(result)



from data_to_paper.terminate.resource_checking import resource_checking
from .latex_doc import LatexDocument


@resource_checking("Checking pdflatex installation")
def check_pdflatex_is_installed():
    # This will raise MissingInstallationError
    LatexDocument().raise_if_pdflatex_is_not_installed()


@resource_checking("Checking pdflatex packages installation")
def check_pdflatex_packages_are_installed():
    # This will raise MissingInstallationError
    LatexDocument().raise_if_packages_are_not_installed()


def check_all_pdflatex_dependencies():
    check_pdflatex_is_installed()
    check_pdflatex_packages_are_installed()



from dataclasses import dataclass


from .base_products_conversers import ReviewBackgroundProductsConverser
from .dual_converser import QuotedReviewDialogDualConverserGPT


@dataclass
class BaseProductsQuotedReviewGPT(QuotedReviewDialogDualConverserGPT, ReviewBackgroundProductsConverser):
    """
    Base class for conversers that specify prior products and then set a goal for the new product
    to be suggested and reviewed.
    The goal is requested from the user as a triple-backtick text thereby allowing extraction.
    Option for reviewing the sections (set max_review_turns > 0).
    """

    def __post_init__(self):
        ReviewBackgroundProductsConverser.__post_init__(self)
        QuotedReviewDialogDualConverserGPT.__post_init__(self)



from dataclasses import dataclass
from typing import List, Dict, Any

from data_to_paper.base_products.product import ValueProduct, Product
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.research_types.hypothesis_testing.scientific_stage import ScientificStage
from data_to_paper.servers.custom_types import Citation


@dataclass
class GoalAndHypothesisProduct(ValueProduct):
    name: str = "Goal and Hypothesis"
    stage: ScientificStage = ScientificStage.GOAL
    value: str = None

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        return ('\n' + self.value).replace('\n# ', '\n' + '#' * (level + 1) + ' ').strip()


@dataclass
class MostSimilarPapersProduct(ValueProduct):
    name: str = "Papers Most Similar to our Research Goal"
    stage: ScientificStage = ScientificStage.ASSESS_NOVELTY
    value: List[Citation] = None

    def _get_citations(self, is_html=False):
        return '\n'.join(citation.pretty_repr(
            fields=('bibtex_id', 'title', 'journal_and_year', 'tldr'),
            is_html=is_html,
        ) for citation in self.value)

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        return self._get_citations(is_html=False)

    def _get_content_as_html(self, level: int, **kwargs):
        return self._get_citations(is_html=True)


@dataclass
class NoveltyAssessmentProduct(ValueProduct):
    name: str = "Novelty Assessment"
    stage: ScientificStage = ScientificStage.ASSESS_NOVELTY
    value: Dict[str, Any] = None

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        results = self.value
        level_str = '#' * (level + 1)
        s = ''
        s += f'{level_str} Similarities:\n'
        for similarity in results["similarities"]:
            s += f'- {similarity}\n'
        s += f'{level_str} Differences:\n'
        for difference in results["differences"]:
            s += f'- {difference}\n'
        s += f'{level_str} Choice:\n- {results["choice"]}\n'
        s += f'{level_str} Explanation:\n- {results["explanation"]}\n'
        return s


@dataclass
class HypothesisTestingPlanProduct(ValueProduct):
    name: str = 'Hypothesis Testing Plan'
    stage: ScientificStage = ScientificStage.PLAN
    value: Dict[str, Dict[str, str]] = None

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        s = ''
        issues = self.value['ISSUES']
        hypotheses = self.value['HYPOTHESES']
        s += f'{"#" * (level + 1)} Statistical considerations:\n'
        for issue, description in issues.items():
            s += f'{"#" * (level + 2)} {issue}:\n{description}\n'

        s += '\n'
        s += f'{"#" * (level + 1)} Hypotheses:\n'
        for hypothesis, test in hypotheses.items():
            s += f'{"#" * (level + 2)} Hypothesis:\n{hypothesis}\n'
            s += f'{"#" * (level + 2)} Test:\n{test}\n\n'
        return s


@dataclass
class NoveltySummaryProduct(Product):
    name: str = 'Assessment of Research Goal Novelty'
    stage: ScientificStage = ScientificStage.ASSESS_NOVELTY
    most_similar_papers: MostSimilarPapersProduct = None
    novelty_assessment: NoveltyAssessmentProduct = None

    def is_valid(self):
        return self.most_similar_papers and self.novelty_assessment

    def _get_content_as_html(self, level: int, **kwargs):
        s = ''
        s += self.most_similar_papers.as_html(level + 1)
        s += '<br>'
        s += self.novelty_assessment.as_html(level + 1)
        return s

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        s = ''
        s += self.most_similar_papers.as_formatted_text(level + 1, view_purpose=view_purpose)
        s += '\n'
        s += self.novelty_assessment.as_formatted_text(level + 1, view_purpose=view_purpose)
        return s



import re


def round_floats(text, target_precision=4, source_precision=10, pad_with_spaces=True):
    def replacer(match):
        num_str = match.group(0)
        if 'e' in num_str:
            before_exp, _ = num_str.split('e')
        else:
            before_exp = num_str

        num_digits = len(before_exp) - 1 - before_exp.startswith('-') - before_exp.startswith('+')
        if num_digits < source_precision:
            return num_str

        value = float(num_str)
        formatted_str = f'{value:.{target_precision}g}'

        # pad with spaces to match the length of the original string
        if pad_with_spaces and len(formatted_str) < len(num_str):
            formatted_str = formatted_str.ljust(len(num_str))
        return formatted_str

    # This regex matches any floating point number or number in scientific notation
    pattern = r'[-+]?[0-9]*\.[0-9]+([eE][-+]?[0-9]+)?'
    return re.sub(pattern, replacer, text)



from functools import partial

import pandas as pd

from itertools import zip_longest
from typing import Callable


def interleave(*iterators):
    """
    interleave([1, 2, 3], [4, 5, 6]) -> [1, 4, 2, 5, 3, 6]
    interleave([1, 2, 3], [4, 5], [7])) -> [1, 4, 7, 2, 5, 3]
    """
    sentinel = object()
    for tuple in zip_longest(*iterators, fillvalue=sentinel):
        for item in tuple:
            if item is not sentinel:
                yield item


def apply_deeply(obj, func: Callable, should_apply: Callable = None):
    """
    Apply a function to all elements of a nested object.
    """
    if isinstance(obj, dict):
        return {key: apply_deeply(value, func, should_apply) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple, set)):
        return type(obj)(apply_deeply(value, func, should_apply) for value in obj)
    elif isinstance(obj, pd.DataFrame):
        return obj.applymap(partial(apply_deeply, func=func, should_apply=should_apply))
    elif isinstance(obj, pd.Series):
        return obj.apply(partial(apply_deeply, func=func, should_apply=should_apply))
    elif should_apply is None or should_apply(obj):
        return func(obj)
    else:
        return obj



from typing import List

from data_to_paper.text.text_extractors import extract_text_between_tags

from .exceptions import FailedToExtractLatexContent
from .latex_section_tags import get_list_of_tag_pairs_for_section_or_fragment

# ChatGPT sometimes adds these to the end of the latex, so we remove them.
REMOVE_FROM_END = [r'\end{document}']


def extract_latex_section_from_response(response: str, section_or_fragment: str, keep_tags: bool = True,
                                        remove_from_end: List[str] = None):
    """
    Extract specified latex part from the LLM response.
    Report errors if the latex part is not found or is empty.
    """
    if remove_from_end is None:
        remove_from_end = REMOVE_FROM_END
    list_of_tag_pairs = get_list_of_tag_pairs_for_section_or_fragment(section_or_fragment)
    for tag_pair in list_of_tag_pairs:
        try:
            latex = extract_text_between_tags(response, *tag_pair, keep_tags=keep_tags, case_sensitive=False)
            break
        except ValueError:
            pass
    else:
        raise FailedToExtractLatexContent(f'Failed to extract {section_or_fragment} from response. '
                                          f'The response does not contain: {list_of_tag_pairs[0]}.')

    latex = latex.strip()
    for remove in remove_from_end:
        if latex.endswith(remove):
            latex = latex[:-len(remove)]
    return latex



import numpy as np

from data_to_paper.env import CODING_MODEL_ENGINE, JSON_MODEL_ENGINE, WRITING_MODEL_ENGINE
from data_to_paper.terminate.resource_checking import resource_checking
from data_to_paper.conversation import Message, Role

from data_to_paper.servers.llm_call import OPENAI_SERVER_CALLER
from data_to_paper.servers.semantic_scholar import SEMANTIC_SCHOLAR_SERVER_CALLER, \
    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER

models_to_check = {CODING_MODEL_ENGINE, JSON_MODEL_ENGINE, WRITING_MODEL_ENGINE}


EMBEDDING_PAPER = {
    "paper_id": "",
    "title": "Hospital outbreak of Middle East respiratory syndrome",
    "abstract": "Between April 1 and May 23, 2013, a total of 23 cases of MERS-CoV ...",
}


@resource_checking("Checking Semantic Scholar Citation-Server connection")
def check_semantic_scholar_connection():
    # Calling the server. Will raise ServerErrorException, InvalidAPIKeyError, or MissingAPIKeyError
    responses = SEMANTIC_SCHOLAR_SERVER_CALLER._get_server_response(query='antibiotics', rows=1)
    if len(responses) != 1:
        raise ConnectionError("Incorrect response from Semantic Scholar server.")


@resource_checking("Checking Semantic Scholar Embedding-Server connection")
def check_semantic_scholar_embedding_connection():
    # Calling the server. Will raise ServerErrorException, or MissingAPIKeyError
    responses = SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER._get_server_response(EMBEDDING_PAPER)
    if not isinstance(responses, np.ndarray):
        raise ConnectionError("Incorrect response from Semantic Scholar Embedding server.")


@resource_checking("Checking LLM Servers connection")
def check_llm_servers_connection():
    for model_engine in models_to_check:
        print(f"Checking LLM Server connection for {model_engine}... ", end='')
        # Calling the server. Will raise ServerErrorException, InvalidAPIKeyError, or MissingAPIKeyError
        OPENAI_SERVER_CALLER._get_server_response(
            messages=[Message(role=Role.USER, content='Your favorite color. Answer in one word')],
            model_engine=model_engine)
        print("OK")


def check_all_servers():
    check_semantic_scholar_connection()
    check_semantic_scholar_embedding_connection()
    check_llm_servers_connection()



from pathlib import Path

from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.utils.mutable import Mutable, Flag
from data_to_paper.servers.types import APIKey
from data_to_paper.types import HumanReviewType
from data_to_paper.utils.scholar_server_caller import ScholarServer

BASE_FOLDER = Path(__file__).parent
FOLDER_FOR_RUN = Path(__file__).parent / "temp_run"

""" API KEYS """
# Define API keys. See INSTALL.md for instructions.
OPENAI_API_KEY = APIKey.from_env("OPENAI_API_KEY")
DEEPINFRA_API_KEY = APIKey.from_env("DEEPINFRA_API_KEY")
SEMANTIC_SCHOLAR_API_KEY = APIKey.from_env("SEMANTIC_SCHOLAR_API_KEY")

""" LLM MODELS """
# Choose LLM model engines:
CODING_MODEL_ENGINE = ModelEngine.GPT4o
JSON_MODEL_ENGINE = ModelEngine.GPT4o
WRITING_MODEL_ENGINE = ModelEngine.GPT4o

""" SCHOLAR SERVER """
# Choose the server for the scholar API:
if SEMANTIC_SCHOLAR_API_KEY.key is not None:
    SCHOLAR_SERVER = ScholarServer("Semantic Scholar")
else:
    SCHOLAR_SERVER = ScholarServer("Crossref")


# Use json mode when requesting LLM structured response:
JSON_MODE = True

""" LLM-CREATED CODE """
# Supported packages for LLM code:
SUPPORTED_PACKAGES = ("numpy", "pandas", "scipy", "sklearn")

# max time for code timeout when running LLM-writen code (seconds)
MAX_EXEC_TIME = Mutable(600)

# Round numbers in LLM code output:
NUM_DIGITS_FOR_FLOATS = 4

# Max number of tokens allowed in code output:
MAX_SENSIBLE_OUTPUT_SIZE_TOKENS = Mutable(2500)

# GPT code environment:
TRACK_P_VALUES = Flag(True)  # must be True in the current version

""" MESSAGE FORMATTING: """
# Text width for conversation output:
TEXT_WIDTH = 150

# Whether to present code debugging iterations as code diff or full.
# Defining: compaction_code_diff = num_lines(new_code) - num_lines(code_diff)
# We show code diff if compaction_code_diff > MINIMAL_COMPACTION_TO_SHOW_CODE_DIFF
MINIMAL_COMPACTION_TO_SHOW_CODE_DIFF = Mutable(
    20
)  # Use 0 to always show code diff, or None to always show full code

# Whether to show incomplete code in the output:
HIDE_INCOMPLETE_CODE = Flag(True)

""" PLAYBACK """
# Delay for cache retrieval (for replay to behave as if we are waiting for the server):
DELAY_CODE_RUN_CACHE_RETRIEVAL = Mutable(0.01)  # seconds
DELAY_SERVER_CACHE_RETRIEVAL = Mutable(0.01)  # seconds

# Pause time (in seconds). 0 for no pause; None to wait for Continue button.
PAUSE_AT_RULE_BASED_FEEDBACK = Mutable(None)
PAUSE_AT_LLM_FEEDBACK = Mutable(None)
PAUSE_AT_PROMPT_FOR_LLM_FEEDBACK = Mutable(None)
PAUSE_AFTER_LITERATURE_SEARCH = Mutable(None)
REQUEST_CONTINUE_IN_PLAYBACK = Flag(True)
FAKE_REQUEST_HUMAN_RESPONSE_ON_PLAYBACK = Flag(False)  # For video recording

""" HUMAN CO-PILOTING """
# CHOSEN_APP:
#   'console': console-based interaction
#   'pyside': GUI-based interaction (requires installing PySide6)
#    None: Does not ask for or records human interactions (legacy. not recommended).
# Runs recorded with 'pyside'/'console' can be replayed with either 'pyside'/'console',
# but not with None. Runs recorded with None can be replayed only with None.
CHOSEN_APP = Mutable("pyside")

# Human review:
# NONE - no human review
# LLM_FIRST - LLM review is performed first and sent to human review
# LLM_UPON_REQUEST - LLM review is performed only upon human request
DEFAULT_HUMAN_REVIEW_TYPE = Mutable(HumanReviewType.LLM_UPON_REQUEST)

# AI review. If True, AI review will default to the terminating phrase exceeding the max_reviewing_rounds.
# (Only in effect when DEFAULT_HUMAN_REVIEW_TYPE is not HumanReviewType.NONE)
AUTO_TERMINATE_AI_REVIEW = Flag(False)

HUMAN_NAME = "Human"

""" DEBUGGING """
SHOW_LLM_CONTEXT = Flag(True)
SAVE_INTERMEDIATE_LATEX = Flag(False)
PRINT_CITATIONS = Flag(True)
DEBUG_MODE = Flag(False)
